/* $Id: x86.c,v 1.22 2001/06/10 22:55:18 mikpe Exp $
 * x86 performance-monitoring counters driver.
 *
 * Copyright (C) 1999-2001  Mikael Pettersson
 */
#include <linux/config.h>
#define __NO_VERSION__
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/perfctr.h>

#include <asm/msr.h>

#include "compat.h"
#include "x86_compat.h"
#include "x86_tests.h"

/* Support for lazy evntsel MSR updates. */
struct perfctr_cpu_state {
	struct perfctr_cpu_control control;
} __attribute__((__aligned__(SMP_CACHE_BYTES)));
static struct perfctr_cpu_state per_cpu_state[NR_CPUS] __cacheline_aligned;

/* Intel P5, Cyrix 6x86MX/MII/III, Centaur WinChip C6/2/3 */
#define MSR_P5_CESR		0x11
#define MSR_P5_CTR0		0x12		/* .. 0x13 */
#define P5_CESR_CPL		0x00C0
#define P5_CESR_RESERVED	(~0x01FF)
#define MII_CESR_RESERVED	(~0x05FF)
#define C6_CESR_RESERVED	(~0x00FF)

/* Intel P6, VIA Cyrix III / C3 */
#define MSR_P6_PERFCTR0		0xC1		/* .. 0xC2 */
#define MSR_P6_EVNTSEL0		0x186		/* .. 0x187 */
#define P6_EVNTSEL_ENABLE	0x00400000
#define P6_EVNTSEL_INT		0x00100000
#define P6_EVNTSEL_CPL		0x00030000
#define P6_EVNTSEL_RESERVED	0x00280000
#define VC3_EVNTSEL1_RESERVED	(~0x1FF)

/* AMD K7 */
#define MSR_K7_EVNTSEL0		0xC0010000	/* .. 0xC0010003 */
#define MSR_K7_PERFCTR0		0xC0010004	/* .. 0xC0010007 */

#define rdmsrl(msr,low) \
	__asm__ __volatile__("rdmsr" : "=a"(low) : "c"(msr) : "edx")
#define rdpmcl(ctr,low) \
	__asm__ __volatile__("rdpmc" : "=a"(low) : "c"(ctr) : "edx")

static unsigned int new_id(void)
{
	static spinlock_t lock = SPIN_LOCK_UNLOCKED;
	static unsigned int counter;
	int id;

	spin_lock(&lock);
	id = ++counter;
	spin_unlock(&lock);
	return id;
}

/****************************************************************
 *								*
 * Driver procedures.						*
 *								*
 ****************************************************************/

/*
 * Intel P5 family (Pentium, family code 5).
 * - One TSC and two 40-bit PMCs.
 * - A single 32-bit CESR (MSR 0x11) controls both PMCs.
 *   CESR has two halves, each controlling one PMC.
 *   To keep the API reasonably clean, the user puts 16 bits of
 *   control data in each counter's evntsel; the driver combines
 *   these to a single 32-bit CESR value.
 * - Overflow interrupts are not available.
 * - Pentium MMX added the RDPMC instruction. RDPMC has lower
 *   overhead than RDMSR and it can be used in user-mode code.
 * - The MMX events are not symmetric: some events are only available
 *   for some PMC, and some event codes denote different events
 *   depending on which PMCs they control.
 * - pmc_map[] is not required to be the identity function.
 */

/* shared with MII and C6 */
static int p5_like_check_control(struct perfctr_cpu_control *control,
				 unsigned int reserved_bits, int is_c6)
{
	unsigned short cesr_half[2];
	unsigned int pmc, evntsel, i;

	if( control->user.nrictrs != 0 || control->user.nractrs > 2 )
		return -EINVAL;
	cesr_half[0] = 0;
	cesr_half[1] = 0;
	for(i = 0; i < control->user.nractrs; ++i) {
		pmc = control->user.pmc_map[i];
		/* pmc_map[] should be the identity on C6 */
		if( is_c6 && pmc != i )
			return -EINVAL;
		if( pmc > 1 || cesr_half[pmc] != 0 )
			return -EINVAL;
		evntsel = control->user.evntsel[i];
		/* protect reserved bits */
		if( (evntsel & reserved_bits) != 0 )
			return -EINVAL;
		/* the CPL field (if defined) must be non-zero */
		if( !is_c6 && !(evntsel & P5_CESR_CPL) )
			return -EINVAL;
		cesr_half[pmc] = evntsel;
	}
	control->k1.p5_cesr = (cesr_half[1] << 16) | cesr_half[0];
	return 0;
}

static int p5_check_control(struct perfctr_cpu_control *control)
{
	return p5_like_check_control(control, P5_CESR_RESERVED, 0);
}

/* shared with MII but not C6 */
static void p5_write_control(const struct perfctr_cpu_control *control)
{
	struct perfctr_cpu_control *cpu;
	unsigned cesr;

	cesr = control->k1.p5_cesr;
	if( !cesr )	/* no PMC is on (this test doesn't work on C6) */
		return;
	cpu = &per_cpu_state[smp_processor_id()].control;
	if( cpu->k1.p5_cesr != cesr ) {
		cpu->k1.p5_cesr = cesr;
		wrmsr(MSR_P5_CESR, cesr, 0);
	}
}

static void p5_read_counters(const struct perfctr_cpu_control *control,
			     struct perfctr_low_ctrs *ctrs)
{
	unsigned int cstatus, nrctrs, i;

	cstatus = control->cstatus;
	if( perfctr_cstatus_has_tsc(cstatus) )
		rdtscl(ctrs->tsc);
	nrctrs = perfctr_cstatus_nrctrs(cstatus);
	for(i = 0; i < nrctrs; ++i) {
		unsigned int pmc = control->user.pmc_map[i];
		rdmsrl(MSR_P5_CTR0+pmc, ctrs->pmc[i]);
	}
}

/* shared with MII but not C6 */
static void p5mmx_read_counters(const struct perfctr_cpu_control *control,
				struct perfctr_low_ctrs *ctrs)
{
	unsigned int cstatus, nrctrs, i;

	cstatus = control->cstatus;
	if( perfctr_cstatus_has_tsc(cstatus) )
		rdtscl(ctrs->tsc);
	nrctrs = perfctr_cstatus_nrctrs(cstatus);
	for(i = 0; i < nrctrs; ++i) {
		unsigned int pmc = control->user.pmc_map[i];
		rdpmcl(pmc, ctrs->pmc[i]);
	}
}

/* shared with MII and C6 */
static void p5_clear_counters(void)
{
	wrmsr(MSR_P5_CESR, 0, 0);
	wrmsr(MSR_P5_CTR0+0, 0, 0);
	wrmsr(MSR_P5_CTR0+1, 0, 0);
}

/*
 * Cyrix 6x86/MII/III.
 * - Same MSR assignments as P5 MMX. Has RDPMC and two 48-bit PMCs.
 * - Event codes and CESR formatting as in the plain P5 subset.
 * - Many but not all P5 MMX event codes are implemented.
 * - Cyrix adds a few more event codes. The event code is widened
 *   to 7 bits, and Cyrix puts the high bit in CESR bit 10
 *   (and CESR bit 26 for PMC1).
 */

static int mii_check_control(struct perfctr_cpu_control *control)
{
	return p5_like_check_control(control, MII_CESR_RESERVED, 0);
}

/*
 * Centaur WinChip C6/2/3.
 * - Same MSR assignments as P5 MMX. Has RDPMC and two 40-bit PMCs.
 * - CESR is formatted with two halves, like P5. However, there
 *   are no defined control fields for e.g. CPL selection, and
 *   there is no defined method for stopping the counters.
 * - Only a few event codes are defined.
 * - The 64-bit TSC is synthesised from the low 32 bits of the
 *   two PMCs, and CESR has to be set up appropriately.
 *   Reprogramming CESR causes RDTSC to yield invalid results.
 *   (The C6 may also hang in this case, due to C6 erratum I-13.)
 *   Therefore, using the PMCs on any of these processors requires
 *   that the TSC is not accessed at all. Because of this side-effect,
 *   WinChip support requires a specific configuration option.
 * - The WinChip event codes and PMCs are symmetric, so pmc_map[] is
 *   required to be identity function.
 * - c6_read_counters() "fakes" a TSC, and c6_check_control allows tsc_on.
 *   This is done to enable user-space to sample without issuing system
 *   calls, using the TSC start field as a context-switch counter.
 */

#if defined(CONFIG_PERFCTR_WINCHIP)
static int c6_check_control(struct perfctr_cpu_control *control)
{
	return p5_like_check_control(control, C6_CESR_RESERVED, 1);
}

static void c6_write_control(const struct perfctr_cpu_control *control)
{
	struct perfctr_cpu_control *cpu;
	unsigned cesr;

	if( control->user.nractrs == 0 )	/* no PMC is on */
		return;
	cpu = &per_cpu_state[smp_processor_id()].control;
	cesr = control->k1.p5_cesr;
	if( cpu->k1.p5_cesr != cesr ) {
		cpu->k1.p5_cesr = cesr;
		wrmsr(MSR_P5_CESR, cesr, 0);
	}
}

static void c6_read_counters(const struct perfctr_cpu_control *control,
			     struct perfctr_low_ctrs *ctrs)
{
	static unsigned int fake_tsc;
	unsigned int nrctrs, i;

	ctrs->tsc = ++fake_tsc;
	nrctrs = perfctr_cstatus_nrctrs(control->cstatus);
	for(i = 0; i < nrctrs; ++i)
		rdpmcl(i, ctrs->pmc[i]);
}
#endif	/* CONFIG_PERFCTR_WINCHIP */

/*
 * Intel P6 family (Pentium Pro, Pentium II, and Pentium III cores,
 * and Xeon and Celeron versions of Pentium II and III cores).
 * - One TSC and two 40-bit PMCs.
 * - One 32-bit EVNTSEL MSR for each PMC.
 * - EVNTSEL0 contains a global enable/disable bit.
 *   That bit is reserved in EVNTSEL1.
 * - Each EVNTSEL contains a CPL field.
 * - Overflow interrupts are possible, but requires that the
 *   local APIC is available. Mobile P6 CPUs have no local APIC.
 *   Additional kernel patches are also required.
 * - The PMCs cannot be initialised with arbitrary values, since
 *   wrmsr fills the high bits by sign-extending from bit 31.
 * - Most events are symmetric, but a few are not.
 * - pmc_map[] is required to be the identity function. PMC1 cannot
 *   be used if PMC0 is skipped (since EVNTSEL0 has the global
 *   enable bit), so the counters might as well be listed in the
 *   natural order.
 */

/* shared with K7 */
static int p6_like_check_control(struct perfctr_cpu_control *control, int is_k7)
{
	unsigned int evntsel, i, nractrs, nrctrs;

	nractrs = control->user.nractrs;
	nrctrs = nractrs + control->user.nrictrs;
	if( nrctrs < nractrs || nrctrs > (is_k7 ? 4 : 2) )
		return -EINVAL;

	/* disallow i-mode counters if we cannot catch the interrupts */
	if( control->user.nrictrs
	    && !(perfctr_info.cpu_features & PERFCTR_FEATURE_PCINT) )
		return -EPERM;

	for(i = 0; i < nrctrs; ++i) {
		/* pmc_map[] should be the identity function */
		if( control->user.pmc_map[i] != i )
			return -EINVAL;
		evntsel = control->user.evntsel[i];
		/* protect reserved bits */
		if( evntsel & P6_EVNTSEL_RESERVED )
			return -EPERM;
		/* check ENable bit */
		if( is_k7 ) {
			/* ENable bit must be set in each evntsel */
			if( !(evntsel & P6_EVNTSEL_ENABLE) )
				return -EPERM;
		} else {
			/* only evntsel[0] has the ENable bit */
			if( evntsel & P6_EVNTSEL_ENABLE ) {
				if( i > 0 )
					return -EPERM;
			} else {
				if( i == 0 )
					return -EINVAL;
			}
		}
		/* the CPL field must be non-zero */
		if( !(evntsel & P6_EVNTSEL_CPL) )
			return -EPERM;
		/* INT bit must be off for a-mode and on for i-mode counters */
		if( evntsel & P6_EVNTSEL_INT ) {
			if( i < nractrs )
				return -EPERM;
		} else {
			if( i >= nractrs )
				return -EPERM;
		}
	}
	control->k1.id = new_id();
	return 0;
}

static int p6_check_control(struct perfctr_cpu_control *control)
{
	return p6_like_check_control(control, 0);
}

static void p6_write_control(const struct perfctr_cpu_control *control)
{
	struct perfctr_cpu_control *cpu;
	unsigned int nrctrs, i;

	cpu = &per_cpu_state[smp_processor_id()].control;
	if( cpu->k1.id == control->k1.id ) {
#ifdef CONFIG_PERFCTR_DEBUG
		nrctrs = perfctr_cstatus_nrctrs(control->cstatus);
		for(i = 0; i < nrctrs; ++i) {
			unsigned int evntsel = control->user.evntsel[i];
			if( evntsel != cpu->user.evntsel[i] ) {
				printk(KERN_ERR __FUNCTION__ ": BUG! evntsel cache inconsistency (pid %d, comm %s)\n",
				       current->pid, current->comm);
				return;
			}
		}
#endif
		return;
	}
	nrctrs = perfctr_cstatus_nrctrs(control->cstatus);
	for(i = 0; i < nrctrs; ++i) {
		unsigned int evntsel = control->user.evntsel[i];
		if( evntsel != cpu->user.evntsel[i] ) {
			cpu->user.evntsel[i] = evntsel;
			wrmsr(MSR_P6_EVNTSEL0+i, evntsel, 0);
		}
	}
	cpu->k1.id = control->k1.id;
}

/* shared with K7 */
static void p6_read_counters(const struct perfctr_cpu_control *control,
			     struct perfctr_low_ctrs *ctrs)
{
	unsigned int cstatus, nrctrs, i;

	cstatus = control->cstatus;
	if( perfctr_cstatus_has_tsc(cstatus) )	/* XXX: ignore this test? */
		rdtscl(ctrs->tsc);
	nrctrs = perfctr_cstatus_nrctrs(cstatus);
	for(i = 0; i < nrctrs; ++i)
		rdpmcl(i, ctrs->pmc[i]);
}

static void p6_clear_counters(void)
{
	int i;

	for(i = 0; i < 2; ++i) {
		wrmsr(MSR_P6_EVNTSEL0+i, 0, 0);
		wrmsr(MSR_P6_PERFCTR0+i, 0, 0);
	}
}

#if PERFCTR_INTERRUPT_SUPPORT
/* PRE: perfctr_cstatus_has_ictrs(control->cstatus) != 0 */
static void p6_isuspend(const struct perfctr_cpu_control *control)
{
	struct perfctr_cpu_control *cpu;

	cpu = &per_cpu_state[smp_processor_id()].control;
	cpu->user.evntsel[0] = 0;
	wrmsr(MSR_P6_EVNTSEL0, 0, 0);
	/* cpu->k1.id is still == control->k1.id */
}

/* PRE: perfctr_cstatus_has_ictrs(control->cstatus) != 0 */
static void p6_iresume(const struct perfctr_cpu_control *control,
		       const struct perfctr_low_ctrs *ctrs)
{
	struct perfctr_cpu_control *cpu;
	unsigned int cstatus, nrctrs, i;

	cpu = &per_cpu_state[smp_processor_id()].control;
	/* XXX: move k1.id test up here? */
	if( cpu->user.evntsel[0] ) {
		cpu->user.evntsel[0] = 0;
		wrmsr(MSR_P6_EVNTSEL0, 0, 0);
		cpu->k1.id = 0;
	} else if( cpu->k1.id == control->k1.id ) {
		/* isuspend() cleared EVNTSEL0, so invalidate the cache
		   here to force write_control() to reload the EVNTSELs.
		   The k1.id cache still allows us to avoid reloading
		   the PERFCTRs. */
		cpu->k1.id = 0;
		return;
	}
	cstatus = control->cstatus;
	nrctrs = perfctr_cstatus_nrctrs(cstatus);
	for(i = perfctr_cstatus_nractrs(cstatus); i < nrctrs; ++i)
		wrmsr(MSR_P6_PERFCTR0+i, ctrs->pmc[i], 0);
	/* cpu->k1.id remains != control->k1.id */
}
#endif	/* PERFCTR_INTERRUPT_SUPPORT */

/*
 * AMD K7 family (Athlon, Duron).
 * - Somewhat similar to the Intel P6 family.
 * - Four 48-bit PMCs.
 * - Four 32-bit EVNTSEL MSRs with similar layout as in P6.
 * - Completely different MSR assignments :-(
 * - Fewer countable events defined :-(
 * - The events appear to be completely symmetric.
 * - The EVNTSEL MSRs are symmetric since each has its own enable bit.
 * - Publicly available documentation is incomplete and incorrect.
 * - pmc_map[] is required to be the identity function.
 */

static int k7_check_control(struct perfctr_cpu_control *control)
{
	return p6_like_check_control(control, 1);
}

static void k7_write_control(const struct perfctr_cpu_control *control)
{
	struct perfctr_cpu_control *cpu;
	unsigned int nrctrs, i;

	cpu = &per_cpu_state[smp_processor_id()].control;
	if( cpu->k1.id == control->k1.id ) {
#ifdef CONFIG_PERFCTR_DEBUG
		nrctrs = perfctr_cstatus_nrctrs(control->cstatus);
		for(i = 0; i < nrctrs; ++i) {
			unsigned int evntsel = control->user.evntsel[i];
			if( evntsel != cpu->user.evntsel[i] ) {
				printk(KERN_ERR __FUNCTION__ ": BUG! evntsel cache inconsistency (pid %d, comm %s)\n",
				       current->pid, current->comm);
				return;
			}
		}
#endif
		return;
	}
	nrctrs = perfctr_cstatus_nrctrs(control->cstatus);
	for(i = 0; i < nrctrs; ++i) {
		unsigned int evntsel = control->user.evntsel[i];
		if( evntsel != cpu->user.evntsel[i] ) {
			cpu->user.evntsel[i] = evntsel;
			wrmsr(MSR_K7_EVNTSEL0+i, evntsel, 0);
		}
	}
	cpu->k1.id = control->k1.id;
}

static void k7_clear_counters(void)
{
	int i;

	for(i = 0; i < 4; ++i) {
		wrmsr(MSR_K7_EVNTSEL0+i, 0, 0);
		wrmsr(MSR_K7_PERFCTR0+i, 0, 0);
	}
}

#if PERFCTR_INTERRUPT_SUPPORT
/* PRE: perfctr_cstatus_has_ictrs(control->cstatus) != 0 */
static void k7_isuspend(const struct perfctr_cpu_control *control)
{
	struct perfctr_cpu_control *cpu;
	unsigned int cstatus, nrctrs, i;

	cpu = &per_cpu_state[smp_processor_id()].control;
	cstatus = control->cstatus;
	nrctrs = perfctr_cstatus_nrctrs(cstatus);
	for(i = perfctr_cstatus_nractrs(cstatus); i < nrctrs; ++i) {
		cpu->user.evntsel[i] = 0;
		wrmsr(MSR_K7_EVNTSEL0+i, 0, 0);
	}
	/* cpu->k1.id is still == control->k1.id */
}

/* PRE: perfctr_cstatus_has_ictrs(control->cstatus) != 0 */
static void k7_iresume(const struct perfctr_cpu_control *control,
		       const struct perfctr_low_ctrs *ctrs)
{
	struct perfctr_cpu_control *cpu;
	unsigned int cstatus, nrctrs, nractrs, i;
	int id_valid;

	cpu = &per_cpu_state[smp_processor_id()].control;
	id_valid = 1;
	cstatus = control->cstatus;
	nrctrs = perfctr_cstatus_nrctrs(cstatus);
	nractrs = perfctr_cstatus_nractrs(cstatus);
	/* XXX: move k1.id test up here? */
	for(i = nractrs; i < nrctrs; ++i) {
		if( cpu->user.evntsel[i] ) {
			cpu->user.evntsel[i] = 0;
			wrmsr(MSR_P6_EVNTSEL0+i, 0, 0);
			id_valid = 0;
		}
	}
	if( !id_valid )
		cpu->k1.id = 0;
	else if( cpu->k1.id == control->k1.id ) {
		cpu->k1.id = 0; /* see comment in p6_iresume() */
		return;
	}
	for(i = nractrs; i < nrctrs; ++i)
		wrmsr(MSR_K7_PERFCTR0+i, ctrs->pmc[i], -1);
	/* cpu->k1.id remains != control->k1.id */
}
#endif	/* PERFCTR_INTERRUPT_SUPPORT */

/*
 * VIA Cyrix III and C3 family.
 * - A Centaur design somewhat similar to the Intel P6.
 * - PERFCTR0 is an alias for the TSC, and EVNTSEL0 is read-only.
 * - PERFCTR1 is 32 bits wide.
 * - EVNTSEL1 has no defined control fields, and there is no
 *   defined method for stopping the counter.
 * - It is not known if the reserved fields in EVNTSEL1 should be
 *   preserved or written with zeroes.
 * - Only a few event codes are defined.
 * - No local APIC or interrupt-mode support.
 * - pmc_map[] is NOT the identity function: pmc_map[0] must be 1,
 *   if nrctrs == 1.
 */
static int vc3_check_control(struct perfctr_cpu_control *control)
{
	if( control->user.nrictrs || control->user.nractrs > 1 )
		return -EINVAL;
	if( control->user.nractrs == 1 ) {
		if( control->user.pmc_map[0] != 1 )
			return -EINVAL;
		if( control->user.evntsel[0] & VC3_EVNTSEL1_RESERVED )
			return -EPERM;
	}
	return 0;
}

static void vc3_write_control(const struct perfctr_cpu_control *control)
{
	struct perfctr_cpu_control *cpu;
	unsigned evntsel;

	if( control->user.nractrs == 0 )	/* PERFCTR1 is not on */
		return;
	cpu = &per_cpu_state[smp_processor_id()].control;
	evntsel = control->user.evntsel[0];
	if( cpu->user.evntsel[0] != evntsel ) {
		/* XXX: read-modify-write to preserve reserved bits */
		unsigned int old, dummy;
		cpu->user.evntsel[0] = evntsel;
		rdmsr(MSR_P6_EVNTSEL0+1, old, dummy);
		evntsel |= (old & VC3_EVNTSEL1_RESERVED);
		wrmsr(MSR_P6_EVNTSEL0+1, evntsel, 0);
	}
}

static void vc3_read_counters(const struct perfctr_cpu_control *control,
			      struct perfctr_low_ctrs *ctrs)
{
	unsigned int cstatus;

	cstatus = control->cstatus;
	if( perfctr_cstatus_has_tsc(cstatus) ) /* XXX: ignore this test? */
		rdtscl(ctrs->tsc);
	if( perfctr_cstatus_nrctrs(cstatus) )
		rdpmcl(1, ctrs->pmc[0]);
}

/*
 * Generic driver for any x86 with a working TSC.
 */

static int generic_check_control(struct perfctr_cpu_control *control)
{
	if( control->user.nractrs || control->user.nrictrs )
		return -EINVAL;
	return 0;
}

static void generic_write_control(const struct perfctr_cpu_control *control)
{
}

static void generic_read_counters(const struct perfctr_cpu_control *control,
				  struct perfctr_low_ctrs *ctrs)
{
	rdtscl(ctrs->tsc);
}

static void generic_clear_counters(void)
{
}

/*
 * Exported driver methods.
 * We export these as named functions, not function pointers.
 * Calls to the high-frequency ones (write_control, read_counters,
 * isuspend and iresume) are back-patched to invoke the correct
 * processor-specific methods directly, thereby saving the
 * overheads of indirect function calls.
 * We don't bother back-patching calls to infrequently used methods.
 */

static void redirect_call(void *ra, void *to)
{
	/* we can only redirect `call near relative' instructions */
	if( *((unsigned char*)ra - 5) != 0xE8 ) {
		printk(KERN_WARNING __FILE__ ":" __FUNCTION__
		       ": unable to redirect caller %p to %p\n",
		       ra, to);
		return;
	}
	*(int*)((char*)ra - 4) = (char*)to - (char*)ra;
}

static int (*check_control)(struct perfctr_cpu_control*);
int perfctr_cpu_check_control(struct perfctr_cpu_control *control)
{
	int ret = check_control(control);
	if( ret == 0 )
		control->cstatus = perfctr_mk_cstatus(control->user.tsc_on,
						      control->user.nractrs,
						      control->user.nrictrs);
	return ret;
}

static void (*write_control)(const struct perfctr_cpu_control*);
void perfctr_cpu_write_control(const struct perfctr_cpu_control *control)
{
	redirect_call(__builtin_return_address(0), write_control);
	return write_control(control);
}

static void (*read_counters)(const struct perfctr_cpu_control *,
			     struct perfctr_low_ctrs*);
void perfctr_cpu_read_counters(const struct perfctr_cpu_control *control,
			       struct perfctr_low_ctrs *ctrs)
{
	redirect_call(__builtin_return_address(0), read_counters);
	return read_counters(control, ctrs);
}

static void (*clear_counters)(void);
static void perfctr_cpu_clear_counters(void)
{
	return clear_counters();
}

#if PERFCTR_INTERRUPT_SUPPORT
static void (*cpu_isuspend)(const struct perfctr_cpu_control*);
void perfctr_cpu_isuspend(const struct perfctr_cpu_control *control)
{
	redirect_call(__builtin_return_address(0), cpu_isuspend);
	return cpu_isuspend(control);
}

static void (*cpu_iresume)(const struct perfctr_cpu_control*,
			   const struct perfctr_low_ctrs*);
void perfctr_cpu_iresume(const struct perfctr_cpu_control *control,
			 const struct perfctr_low_ctrs *ctrs)
{
	redirect_call(__builtin_return_address(0), cpu_iresume);
	return cpu_iresume(control, ctrs);
}
#endif	/* PERFCTR_INTERRUPT_SUPPORT */

/****************************************************************
 *								*
 * Processor detection and initialisation procedures.		*
 *								*
 ****************************************************************/

static int __init intel_init(void)
{
	if( !cpu_has_tsc )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 5:
		if( cpu_has_mmx ) {
			perfctr_info.cpu_type = PERFCTR_X86_INTEL_P5MMX;
			read_counters = p5mmx_read_counters;
		} else {
			perfctr_info.cpu_type = PERFCTR_X86_INTEL_P5;
			perfctr_info.cpu_features &= ~PERFCTR_FEATURE_RDPMC;
			read_counters = p5_read_counters;
		}
		write_control = p5_write_control;
		check_control = p5_check_control;
		clear_counters = p5_clear_counters;
		perfctr_p5_init_tests();
		return 0;
	case 6:
		if( boot_cpu_data.x86_model >= 7 )	/* PIII */
			perfctr_info.cpu_type = PERFCTR_X86_INTEL_PIII;
		else if( boot_cpu_data.x86_model >= 3 )	/* PII or Celeron */
			perfctr_info.cpu_type = PERFCTR_X86_INTEL_PII;
		else
			perfctr_info.cpu_type = PERFCTR_X86_INTEL_P6;
		read_counters = p6_read_counters;
		write_control = p6_write_control;
		check_control = p6_check_control;
		clear_counters = p6_clear_counters;
#if PERFCTR_INTERRUPT_SUPPORT
		if( cpu_has_apic ) {
			perfctr_info.cpu_features |= PERFCTR_FEATURE_PCINT;
			cpu_isuspend = p6_isuspend;
			cpu_iresume = p6_iresume;
		}
#endif
		perfctr_p6_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init amd_init(void)
{
	if( !cpu_has_tsc )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 6:	/* K7. Model 1 does not have a local APIC. */
		perfctr_info.cpu_type = PERFCTR_X86_AMD_K7;
		read_counters = p6_read_counters;
		write_control = k7_write_control;
		check_control = k7_check_control;
		clear_counters = k7_clear_counters;
#if PERFCTR_INTERRUPT_SUPPORT
		if( cpu_has_apic ) {
			perfctr_info.cpu_features |= PERFCTR_FEATURE_PCINT;
			cpu_isuspend = k7_isuspend;
			cpu_iresume = k7_iresume;
		}
#endif
		perfctr_k7_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init cyrix_init(void)
{
	if( !cpu_has_tsc )
		return -ENODEV;
	switch( boot_cpu_data.x86 ) {
	case 6:	/* 6x86MX, MII, or III */
		perfctr_info.cpu_type = PERFCTR_X86_CYRIX_MII;
		read_counters = p5mmx_read_counters;
		write_control = p5_write_control;
		check_control = mii_check_control;
		clear_counters = p5_clear_counters;
		perfctr_mii_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init centaur_init(void)
{
	switch( boot_cpu_data.x86 ) {
#if defined(CONFIG_PERFCTR_WINCHIP)
	case 5:
		switch( boot_cpu_data.x86_model ) {
		case 4: /* WinChip C6 */
			perfctr_info.cpu_type = PERFCTR_X86_WINCHIP_C6;
			break;
		case 8: /* WinChip 2, 2A, or 2B */
		case 9: /* WinChip 3, a 2A with larger cache and lower voltage */
			perfctr_info.cpu_type = PERFCTR_X86_WINCHIP_2;
			break;
		default:
			return -ENODEV;
		}
		/*
		 * TSC must be inaccessible for perfctrs to work.
		 */
		if( !(read_cr4() & X86_CR4_TSD) || cpu_has_tsc )
			return -ENODEV;
		perfctr_info.cpu_features &= ~PERFCTR_FEATURE_RDTSC;
		read_counters = c6_read_counters;
		write_control = c6_write_control;
		check_control = c6_check_control;
		clear_counters = p5_clear_counters;
		perfctr_c6_init_tests();
		return 0;
#endif
	case 6: /* VIA Cyrix III / C3 */
		if( !cpu_has_tsc )
			return -ENODEV;
		switch( boot_cpu_data.x86_model ) {
		case 6:	/* VIA Cyrix III */
		case 7:	/* VIA C3 */
			break;
		default:
			return -ENODEV;
		}
		perfctr_info.cpu_type = PERFCTR_X86_VIA_C3;
		read_counters = vc3_read_counters;
		write_control = vc3_write_control;
		check_control = vc3_check_control;
		clear_counters = generic_clear_counters;
		perfctr_vc3_init_tests();
		return 0;
	}
	return -ENODEV;
}

static int __init generic_init(void)
{
	if( !cpu_has_tsc )
		return -ENODEV;
	perfctr_info.cpu_features &= ~PERFCTR_FEATURE_RDPMC;
	perfctr_info.cpu_type = PERFCTR_X86_GENERIC;
	check_control = generic_check_control;
	write_control = generic_write_control;
	read_counters = generic_read_counters;
	clear_counters = generic_clear_counters;
	return 0;
}

static char generic_name[] __initdata = "Generic x86 with TSC";
static char p5_name[] __initdata = "Intel Pentium";
static char p5mmx_name[] __initdata = "Intel Pentium MMX";
static char p6_name[] __initdata = "Intel Pentium Pro";
static char pii_name[] __initdata = "Intel Pentium II";
static char piii_name[] __initdata = "Intel Pentium III";
static char mii_name[] __initdata = "Cyrix 6x86MX/MII/III";
static char wcc6_name[] __initdata = "WinChip C6";
static char wc2_name[] __initdata = "WinChip 2/3";
static char k7_name[] __initdata = "AMD K7";
static char vc3_name[] __initdata = "VIA Cyrix III / C3";

char *perfctr_cpu_name[] __initdata = {
	[PERFCTR_X86_GENERIC] generic_name,
	[PERFCTR_X86_INTEL_P5] p5_name,
	[PERFCTR_X86_INTEL_P5MMX] p5mmx_name,
	[PERFCTR_X86_INTEL_P6] p6_name,
	[PERFCTR_X86_INTEL_PII] pii_name,
	[PERFCTR_X86_INTEL_PIII] piii_name,
	[PERFCTR_X86_CYRIX_MII] mii_name,
	[PERFCTR_X86_WINCHIP_C6] wcc6_name,
	[PERFCTR_X86_WINCHIP_2] wc2_name,
	[PERFCTR_X86_AMD_K7] k7_name,
	[PERFCTR_X86_VIA_C3] vc3_name,
};

static void __init perfctr_cpu_init_one(void *ignore)
{
	perfctr_cpu_clear_counters();
#if PERFCTR_INTERRUPT_SUPPORT
	if( cpu_has_apic )
		apic_write(APIC_LVTPC, LOCAL_PERFCTR_VECTOR);
#endif
	if( perfctr_info.cpu_features & PERFCTR_FEATURE_RDPMC )
		set_in_cr4_local(X86_CR4_PCE);
}

static void __exit perfctr_cpu_exit_one(void *ignore)
{
	perfctr_cpu_clear_counters();
#if PERFCTR_INTERRUPT_SUPPORT
	if( cpu_has_apic )
		apic_write(APIC_LVTPC, APIC_DM_NMI | APIC_LVT_MASKED);
#endif
	if( perfctr_info.cpu_features & PERFCTR_FEATURE_RDPMC )
		clear_in_cr4_local(X86_CR4_PCE);
}

#if defined(NMI_LOCAL_APIC) && defined(CONFIG_PM)

static void __init unregister_nmi_pmdev(void)
{
	if( nmi_pmdev ) {
		apic_pm_unregister(nmi_pmdev);
		nmi_pmdev = 0;
	}
}

static int x86_pm_callback(struct pm_dev *dev, pm_request_t rqst, void *data)
{
	/* XXX: incomplete */
	return 0;
}

static struct pm_dev *x86_pmdev;

static void __init x86_pm_init(void)
{
	x86_pmdev = apic_pm_register(PM_SYS_DEV, 0, x86_pm_callback);
}

static void __exit x86_pm_exit(void)
{
	if( x86_pmdev ) {
		apic_pm_unregister(x86_pmdev);
		x86_pmdev = NULL;
	}
}

#else

static inline void unregister_nmi_pmdev(void) { }
static inline void x86_pm_init(void) { }
static inline void x86_pm_exit(void) { }

#endif	/* NMI_LOCAL_APIC && CONFIG_PM */

#if defined(NMI_LOCAL_APIC)

static void __init disable_nmi_watchdog(void)
{
	if( nmi_perfctr_msr ) {
		nmi_perfctr_msr = 0;
		printk(KERN_NOTICE "perfctr: disabled nmi_watchdog\n");
		unregister_nmi_pmdev();
	}
}

#else

static inline void disable_nmi_watchdog(void) { }

#endif

int __init perfctr_cpu_init(void)
{
	int err = -ENODEV;

	/* RDPMC and RDTSC are on by default. They will be disabled
	   by the init procedures if necessary. */
	perfctr_info.cpu_features = PERFCTR_FEATURE_RDPMC | PERFCTR_FEATURE_RDTSC;

	if( cpu_has_msr ) {
		switch( boot_cpu_data.x86_vendor ) {
		case X86_VENDOR_INTEL:
			err = intel_init();
			break;
		case X86_VENDOR_AMD:
			err = amd_init();
			break;
		case X86_VENDOR_CYRIX:
			err = cyrix_init();
			break;
		case X86_VENDOR_CENTAUR:
			err = centaur_init();
		}
	}
	if( err ) {
		err = generic_init();	/* last resort */
		if( err )
			return err;
	}
	/*
	 * Put the hardware in a sane state:
	 * - clear perfctr MSRs
	 * - set up APIC_LVTPC
	 * - set CR4.PCE [on permanently due to __flush_tlb_global()]
	 * - install our default interrupt handler
	 */
	if( perfctr_info.cpu_features & PERFCTR_FEATURE_RDPMC )
		mmu_cr4_features |= X86_CR4_PCE;
	perfctr_cpu_init_one(NULL);
	smp_call_function(perfctr_cpu_init_one, NULL, 1, 1);
	perfctr_cpu_set_ihandler(NULL);
	/*
	 * Fix up the connection to the local APIC:
	 * - disable and disconnect the NMI watchdog
	 * - register our PM callback
	 */
	disable_nmi_watchdog();
	x86_pm_init();
	/*
	 * per_cpu_state[] is initialised to contain "impossible"
	 * evntsel values guaranteed to differ from anything accepted
	 * by perfctr_cpu_check_control(). This way, initialisation of
	 * a CPU's evntsel MSRs will happen automatically the first time
	 * perfctr_cpu_write_control() executes on it.
	 * All-bits-one works for all currently supported processors.
	 * The memset also sets the ids to -1, which is intentional.
	 */
	memset(per_cpu_state, ~0, sizeof per_cpu_state);

	perfctr_info.cpu_khz = cpu_khz;
	perfctr_info.nrcpus = smp_num_cpus;

	return 0;
}

void __exit perfctr_cpu_exit(void)
{
	if( perfctr_info.cpu_features & PERFCTR_FEATURE_RDPMC )
		mmu_cr4_features &= ~X86_CR4_PCE;
	perfctr_cpu_exit_one(NULL);
	smp_call_function(perfctr_cpu_exit_one, NULL, 1, 1);
	perfctr_cpu_set_ihandler(NULL);
	x86_pm_exit();
	/* XXX: restart nmi watchdog? */
}

/****************************************************************
 *								*
 * Hardware reservation.					*
 *								*
 ****************************************************************/

static const char *current_service = 0;

const char *perfctr_cpu_reserve(const char *service)
{
	if( current_service )
		return current_service;
	current_service = service;
	MOD_INC_USE_COUNT;
	return 0;
}

void perfctr_cpu_release(const char *service)
{
	if( service != current_service ) {
		printk(KERN_ERR __FUNCTION__
		       ": attempt by %s to release while reserved by %s\n",
		       service, current_service);
	} else {
		perfctr_cpu_set_ihandler(NULL);
		current_service = 0;
		MOD_DEC_USE_COUNT;
	}
}
